/*-*-C-*-
 * Copyright 2006  Petter Urkedal
 *
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef CUFLOW_WORKERS_H
#define CUFLOW_WORKERS_H

#include <cuflow/fwd.h>
#include <cu/clos.h>
#include <time.h>
#include <unistd.h>

CU_BEGIN_DECLARATIONS
/*!\defgroup cuflow_workers_h cuflow/workers.h: Auxiliary Worker Threads
 *@{\ingroup cuflow_mod
 *
 * This header provides an API to launch one or more worker threads and run
 * functions within them, either as soon as one is idle or at a given time.
 * The main purpose of this is to unify such work in a common framework, so
 * that the same thread(s) can be used, rather than adding now threads for
 * each task.  The functions defined here is aimed to be the bare minimum to
 * achieve this.  For instance, if you need to know when a callback is
 * finised, you may need to supply it's closure struct with an additional
 * mutex and possibly a condition variable to communicate with consumer(s) of
 * the computation.
 *
 * This is implemenated purely with POSIX threads, mutices and condition
 * variables, and in particular we avoid timers or signals which could have
 * caused trouble in a multi-threaded environment.
 */

/*!The current number of worker threads. */
int cuflow_workers_count(void);

/*!Spawn or drops worker threads until there are exactly \a target_count left.
 * This is not safe to call within worker threads. */
void cuflow_workers_spawn(int target_count);

/*!Spawns worker threads as needed so that there are at least \a target_count
 * of them. */
void cuflow_workers_spawn_at_least(int target_count);

/*!Drops worker threads as needed so that there are at most \a target_count
 * left. */
void cuflow_workers_spawn_at_most(int target_count);

/*!Call \a callback in one of the worker threads, as soon as one is ready. */
void cuflow_workers_call(cu_clop0(callback, void));

/*!Call \a callback at absolute \c CLOCK_REALTIME \a t_call. */
void cuflow_workers_call_at(cu_clop(callback, void, struct timespec *t_now), 
			    struct timespec *t_call);
			    

/*!@}*/
CU_END_DECLARATIONS

#endif
